<?php

namespace Drupal\cache_tools\Service;

use Drupal\Core\Cache\CacheTagsInvalidatorInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeInterface;

/**
 * Sanitize cache tags.
 */
class CacheSanitizer {

  /**
   * The cache tag invalidator.
   *
   * @var \Drupal\Core\Cache\CacheTagsInvalidatorInterface
   */
  protected $cacheTagInvalidator;

  /**
   * Cache settings.
   *
   * @var array
   */
  protected $settings;

  /**
   * CacheSanitizer constructor.
   *
   * @param \Drupal\Core\Cache\CacheTagsInvalidatorInterface $cache_tag_invalidator
   *   The cache tag invalidator.
   * @param array $settings
   *   The cache settings.
   */
  public function __construct(CacheTagsInvalidatorInterface $cache_tag_invalidator, array $settings) {
    $this->cacheTagInvalidator = $cache_tag_invalidator;
    $this->settings = $settings;
  }

  /**
   * Filter cacheable metadata based on given options.
   *
   * @param array $filterCacheableMetadata
   *   Cacheable metadata to filter out.
   * @param array $cacheableMetadata
   *   Cacheable metadata.
   * @param array $options
   *   Options to include/exclude cacheable metadata.
   *
   * @return array
   *   Filtered cacheable metadata.
   */
  protected function filterCacheableMetadata(array $filterCacheableMetadata, array $cacheableMetadata, array $options) {
    if (!empty($options['include'])) {
      $cacheableMetadata = array_merge($cacheableMetadata, $options['include']);
    }
    if (!empty($options['exclude'])) {
      $filterCacheableMetadata = array_merge($filterCacheableMetadata, $options['exclude']);
    }
    return array_diff($cacheableMetadata, $filterCacheableMetadata);
  }

  /**
   * Sanitizes cacheable contexts.
   *
   * @param array $cachebleContexts
   *   Cacheable contexts.
   * @param array $options
   *   Optional. List of cache contexts to include/exclude.
   *
   * @return array
   *   Sanitized cacheble contexts.
   */
  public function sanitizeCacheableContexts(array $cachebleContexts, array $options = []) {
    $defaults = $this->settings['sanitize']['defaults']['contexts'] ?? [];
    return $this->filterCacheableMetadata($defaults, $cachebleContexts, $options);
  }

  /**
   * Sanitizes cacheable tags.
   *
   * @param array $cachebleTags
   *   Cacheable contexts.
   * @param array $options
   *   Optional. List of cache tags to include/exclude.
   *
   * @return array
   *   Sanitized cacheble tags.
   */
  public function sanitizeCacheableTags(array $cachebleTags, array $options = []) {
    $defaults = $this->settings['sanitize']['defaults']['tags'] ?? [];
    return $this->filterCacheableMetadata($defaults, $cachebleTags, $options);
  }

  /**
   * Sanitizes cacheable metadata.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity.
   * @param array $build
   *   Build.
   */
  public function sanitize(EntityInterface $entity, array &$build) {
    // Skip if entity is not listed for sanitizing.
    $sanitize = array_key_exists($entity->id(), $this->settings['sanitize'][$entity->getEntityTypeId()]);
    if (!$sanitize) {
      return;
    }
    $cacheSettings = $this->settings['sanitize'][$entity->getEntityTypeId()][$entity->id()] ?? [];
    $this->sanitizeBuild($build, $cacheSettings);
  }

  /**
   * Sanitize build.
   *
   * @param array $build
   *   Build.
   * @param array $options
   *   Optional. Settings to adjust sanitizing.
   */
  public function sanitizeBuild(array &$build, array $options = []) {
    $contextsOptions = $options['contexts'] ?? [];
    $tagsOptions = $options['tags'] ?? [];
    // Filter out undesired cache contexts.
    if (!empty($build['#cache']['contexts'])) {
      $build['#cache']['contexts'] = $this->sanitizeCacheableContexts($build['#cache']['contexts'], $contextsOptions);
    }
    // Filter out undesired cache tags.
    if (!empty($build['#cache']['tags'])) {
      $build['#cache']['tags'] = $this->sanitizeCacheableTags($build['#cache']['tags'], $tagsOptions);
    }
    // Sanitize also build content.
    if (!empty($build['content']['#cache']['contexts'])) {
      $build['content']['#cache']['contexts'] = $this->sanitizeCacheableContexts($build['content']['#cache']['contexts'], $contextsOptions);
    }
    if (!empty($build['content']['#cache']['tags'])) {
      $build['content']['#cache']['tags'] = $this->sanitizeCacheableTags($build['content']['#cache']['tags'], $tagsOptions);
    }
  }

  /**
   * Get published cache tag in format `entitytype_entitybundle_pub`.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity.
   *
   * @return string
   *   Published ache tag.
   */
  public function getPublishedEntityCacheTag(EntityInterface $entity) {
    return $entity->getEntityTypeId() . '_' . $entity->bundle() . '_pub';
  }

  /**
   * Get published cache tag in format `entitytype_pub`.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entityType
   *   Entity.
   *
   * @return string
   *   Published cache tag.
   */
  public function getPublishedEntityTypeCacheTag(EntityTypeInterface $entityType) {
    return $entityType->id() . '_pub';
  }

  /**
   * Invalidates published entity.
   *
   * On entity update where original entity is unpublished and
   * going to be published and on entity insert where new entity is published.
   * Other cases are already covered by other tags:
   * 1. Unpublished entities (created or updated) does not affect anything.
   * 2. Published entities which stay published are invalidated via
   *    entity_type:id (eg node:123).
   * 3. Published entities going to be deleted or unpublished are invalidated
   *    via entity_type:id (eg node:123).
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity.
   *
   * @return bool
   *   True if successful invalidation. False otherwise.
   */
  public function invalidatePublishedEntity(EntityInterface $entity) {
    // Skip if entity type is not allowed by entity type and bundle.
    if (!isset($this->settings['invalidate'][$entity->getEntityTypeId()])) {
      return FALSE;
    }
    if (!in_array($entity->bundle(), $this->settings['invalidate'][$entity->getEntityTypeId()])) {
      return FALSE;
    }
    $tags = [
      $this->getPublishedEntityTypeCacheTag($entity->getEntityType()),
      $this->getPublishedEntityCacheTag($entity),
    ];
    // If this is update we need to take a look at original entity as well.
    $entities = [$entity];
    if ($entity->original) {
      $entities[] = $entity->original;
    }
    // If entity is going to be unpublished (insert)
    // or stays unpublished (update), skip.
    foreach ($entities as $index => $entity) {
      if (empty($entity->get('status')->value)) {
        unset($entities[$index]);
      }
    }
    if ($entities) {
      $this->cacheTagInvalidator->invalidateTags($tags);
      return TRUE;
    }
    return FALSE;
  }

}
